// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: fsearchcgi.cc,v 0.15 1998-01-27 12:30:30+09 harada Exp $
//
#include <iostream.h>
#include <strstream.h>
extern "C" {
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#include <sys/file.h>
}
#include "Retriever.h"
#include "Unifier.h"
#include "Expression.h"
#include "Evaluator.h"
#include "Result.h"
#include "Description.h"

#include "kcnvlib.h"
#include "htmlutil.h"

char param_index[NAME_MAX + 1];
char param_style[16];
int param_from;
int param_n;

#ifdef HAS_FLOCK
int
getLock(int fd){
    return flock(fd, LOCK_EX);
}
int
putLock(int fd){
    return flock(fd, LOCK_UN);
}
#else
static struct flock lock;
int
getLock(int fd){
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // until end of the file
    lock.l_type = F_WRLCK;      // exclusive/write
    lock.l_pid = 0;
    int r;
    while ((r = fcntl(fd, F_SETLKW, &lock)) < 0 && errno == EINTR)
        continue;
    return r < 0 ? r : 0;
}

int
putLock(int fd){
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // until end of the file
    lock.l_type = F_UNLCK;      // unlock
    lock.l_pid = 0;
    int r;
    while ((r = fcntl(fd, F_SETLKW, &lock)) < 0 && errno == EINTR)
        continue;
    return r < 0 ? r : 0;
}
#endif

#define FSEARCHCGI_LOGPATH "fsearch.log"

#ifdef DISABLE_LOGGING
#define writelog(q,r) ;
#else
// log some information into file
// format is:
//   DATE REMOTE_HOST REMOTE_USER INDEXNAME STYLE resultNum from n query\n
int
writelog(const char *query, int resultNum){
    char *host = getenv("REMOTE_HOST");
    if(NULL == host) host = getenv("REMOTE_ADDR");
    if(NULL == host || '\0' == host[0]) host = "-";

    char *user = getenv("REMOTE_USER");
    if(NULL == user || '\0' == user[0]) user = "-";

    char datebuf[64];
    time_t now = time(NULL);
    strftime(datebuf, sizeof(datebuf) - 1,
             "%d %b %Y %H:%M:%S %Z", localtime(&now));

    if(NULL == query || '\0' == query[0]) query = "-";
    ostrstream line;
    line << datebuf << ' ' << host << ' ' << user << ' '
         << param_index << ' ' << param_style << ' '
         << resultNum << ' ' << param_from << ' ' << param_n << ' '
         << query << '\n' << '\0';

    int fd = open(FSEARCHCGI_LOGPATH, O_WRONLY|O_CREAT, 0666);
    if(-1 == fd) return -1;     // exit silently (without logging)
    errno = 0;
    if(0 != getLock(fd)) return -1;

    char *linestr = line.str();
    lseek(fd, 0, SEEK_END);
    write(fd, linestr, strlen(linestr));
    putLock(fd);
    close(fd);
    delete []linestr;

    return errno;
}
#endif

// replace keyword("$foo") as str
int
replace_string(char *buf, const char *keyword, const char *str){
    int replaced = 0;
    int keyLen = strlen(keyword);
    int strLen = strlen(str);
    char *cp = buf;
    while((cp = strstr(cp, keyword))){
        memmove(cp + strLen, cp + keyLen,
                strlen(buf) - (cp + keyLen - buf) + 1);
        strncpy(cp, str, strLen);
        cp += strLen;
        replaced++;
    }
    return replaced;
}

int
display_html(ostream &out, const char *filepath,
             const char *title, const char *query, const char *errormsg,
             int resultNum){
    fstream htmlfile(filepath, ios::in);
    if(! htmlfile.good())
        return -1;

    char qtitle[LINE_MAX];
    char qerrormsg[LINE_MAX];
    char qquery[LINE_MAX];
    html_entitize(qtitle, title);
    html_entitize(qerrormsg, errormsg);
    html_entitize(qquery, query);
    ostrstream numstr, dispnumstr;
    numstr << ' ' << resultNum << '\0';
    char *resultnum = numstr.str();
    dispnumstr << param_n << '\0';
    char *dispnum = dispnumstr.str();

    char linebuf[LINE_MAX];
    for(;;){
        htmlfile.getline(linebuf, LINE_MAX, '\n');
        if(htmlfile.fail()){
            if(! htmlfile.eof())
                PRINTERRMSG("Error occured while printing .html");
            break;
        }

        replace_string(linebuf, "$title", qtitle);
        replace_string(linebuf, "$errormsg", qerrormsg);
        replace_string(linebuf, "$query", qquery);
        replace_string(linebuf, "$index", param_index);
        replace_string(linebuf, "$result", resultnum);
        replace_string(linebuf, "$style", param_style);
        replace_string(linebuf, "$dispnum", dispnum);
        out << linebuf << endl;
    }
    htmlfile.close();
    delete []resultnum;
    delete []dispnum;
    return 0;
}

void
cgi_error(const char *msg1, const char *msg2 = NULL,
          const char *query = NULL, int log = 1){
    if(query == NULL) query = "";
    if(msg2 == NULL) msg2 = "";
    cout << "Content-type: text/html\n\n";
    if(-1 == display_html(cout, "error.html", msg1, query, msg2, 0)){
        cout << "<HTML><HEAD><TITLE>" << msg1
             << "</TITLE></HEAD>\n<BODY><H1>" << msg2
             << "</H1></BODY>\n</HTML>\n";
    }
    if(log) writelog(query, -1);
}

int
get_query(char *key){
    char *method = getenv("REQUEST_METHOD");
    if(! (method && 0 == strcmp(method, "GET"))) {
        if(method){
            cgi_error("Error", "unsupported method");
        }else{
            cgi_error("Error", "no method provided");
        }
        return 1;
    }
    char *qs = getenv("QUERY_STRING");
    if(NULL == qs || '\0' == qs[0]){
        cgi_error("Error", "No query information to decode.");
        return 1;
    }
    char *k, *v;
    while((k = strtok(qs, "&"))){
        qs = NULL;
        v = strchr(k, '=');
        if(v != NULL){
            *v++ = '\0';
            unescape_url(v);
            // cout << "k = \"" << k << "\", param = \"" << param << "\"\n";
            if(0 == strcmp(k, "key")){
                if(NULL == conv2euc(key, v, CODE_UNKNOWN)){
                    strncpy(key, v, MAXQUERYLENGTH);
                }
            }else if(0 == strcmp(k, "from")){
                param_from = atoi(v);
            }else if(0 == strcmp(k, "index")){
                strncpy(param_index, v, sizeof(param_index) - 1);
            }else if(0 == strcmp(k, "style")){
                strncpy(param_style, v, sizeof(param_style) - 1);
            }else if(0 == strcmp(k, "n")){
                param_n = atoi(v);
            }
        }
    }

    return 0;                   // success
}

int
print_results(fstream &descfile, const Result *result, const char *key,
              int resultNum, int disp, int from, int to){
    assert(to >= from);
    assert(to - from < resultNum);

    Description *desc;
    if(0 == strcmp(param_style, "mail"))
        desc = new MailDescription(descfile);
    else if(0 == strcmp(param_style, "html"))
        desc = new HtmlDescription(descfile);
    else
        desc = new Description(descfile); // generic

    int r;
    cout << "<DL>\n";
    for(r = from; r <= to; r++){
        desc->printHtml(cout, result[r]);
    }
    cout << "</DL>" << endl;
    delete desc;
    
    char *cgipath = getenv("SCRIPT_NAME");
    if(NULL == cgipath || '\0' == cgipath[0]){
        cgipath = getenv("REQUEST_URI");
        if(NULL == cgipath || '\0' == cgipath[0])
            cgipath = "fsearch.cgi";
        else{
            char *cp = cgipath;
            while(*cp){
                if(*cp == '?'){ *cp = '\0'; break; }
                cp++;
            }
        }
    }

    char *escaped_key = escape_url(key);

    cout << "\n<BR>\n<BR>\n<SMALL>\n";
    if(from > 0){
        cout << "<A HREF=\"" << cgipath << "?key=" << escaped_key
             << "&from=" << ((from - disp < 0) ? 0 : from - disp);
        cout << "&n=" << disp << "&index=" << param_index  << "\">"
             << "[PREV]</A>" << endl;
    }
    int i, p;
    int i_from = from - 10 * disp;
    if(i_from < 0) i_from = 0;
    int i_to = i_from + 20 * disp;
    if(i_to > resultNum) i_to = resultNum;

    for(i = i_from,p = i / disp; i < i_to; i += disp, p++){
        if(i == from)
            cout << "<B>[" << p << "]</B>" << endl;
        else{
            cout << "<A HREF=\"" << cgipath << "?key=" << escaped_key
                 << "&from=" << i << "&n=" << disp
                 << "&index=" << param_index << "\">"
                 << '[' << p << "]</A>" << endl;
        }
    }
    if(to + 1 < resultNum){
        cout << "<A HREF=\"" << cgipath << "?key=" << escaped_key
             << "&from=" << to + 1 << "&n=" << disp
             << "&index=" << param_index << "\">"
             << "[NEXT]</A>" << endl;
    }
    cout << "</SMALL>\n";
    delete []escaped_key;

    return 0;
}

int
main(){
    char query[MAXQUERYLENGTH + 1];
    param_from = 0;
    param_n = RESULTSFORDISPLAY;
    strcpy(param_index, "default");
    strcpy(param_style, "html");

    query[0] = '\0';
    if(0 != get_query(query)){ // error
        return 0;
    }
    param_from++;
    if(query[0] == '\0'){
        cgi_error("顼: Ĥޤ",
                  "ϤƤ", "");
        return 0;
    }
    if(param_n < 0 || param_n > MAXRESULTSFORDISPLAY){
        param_n = RESULTSFORDISPLAY;
        cgi_error("顼: ϰϤǤ",
                  "n=...ŬڤͤˤƤ", "");
        return 0;
    }

    //
    // search
    //
    char uquery[MAXQUERYLENGTH + 1];
    const Unifier unif;
    unif.unify(uquery, query);
    ComplexExpression cexp(uquery);
    if(cexp.error()){
        cgi_error("顼: Ǥ",
                  cexp.errmsgs[cexp.error()], uquery);
        return 2;
    }
    int resultLen = -1;
    Result *result = NULL;
    // result = checkCache(uquery, resultLen);
    if(-1 == resultLen){        // it didn't hit cache
        char indexpath[PATH_MAX];
        char lexpath[PATH_MAX];
        char xmappath[PATH_MAX];
        strcpy(indexpath, param_index); strcat(indexpath, ".idx");
        strcpy(lexpath, param_index); strcat(lexpath, ".lex");
        strcpy(xmappath, param_index); strcat(xmappath, ".map");

        fstream idxstream(indexpath, ios::in);
        if(! idxstream.good()){
            cgi_error("Error", "cannot open index file.", query);
            return 1;
        }
        IndexFile idxfile(idxstream, 1, 0); // readonly, !newfile
        fstream lexfs(lexpath, ios::in);
        if(! lexfs.good()){
            cgi_error("Error", "cannot open lexicon file.", query);
            return 1;
        }
        LexiconFile lexfile(lexfs, 1, 0); // readonly, !newfile

        int fd = open(xmappath, O_RDONLY, 0666);
        if(fd == -1){
            cgi_error("Error", "cannot open map file", query);
            return 1;
        }

        // setup retriever object
        XMap map(fd);
        Retriever retr(lexfile, idxfile, map);
        Evaluator evaluator;           // use standard ranking policy

        result = cexp.eval(retr, evaluator, resultLen);
        if(resultLen > 0) evaluator.rank(result, resultLen);
        // saveCache(result, resultLen);
    }

    //
    // display results
    //
    if(resultLen <= 0){
        resultLen = 0;
        cgi_error("ĤޤǤ", "nothing found. ", uquery, 0);
        writelog(uquery, 0);
    }else{
        cout << "Content-type: text/html\n\n";

        char descpath[PATH_MAX];
        strcpy(descpath, param_index); strcat(descpath, ".dsc");
        fstream descfile(descpath, ios::in);
        if(! descfile.good()){
            cgi_error("Error", "cannot open description file.", uquery);
            return 3;
        }

        if(param_from <= 0 || param_from > resultLen){
            cgi_error("Error", "parameter 'from' out of range.", uquery);
            return 4;
        }
        int rto = param_from + param_n - 1;
        if(rto > resultLen) rto = resultLen;
        if(rto == 0)
            rto = resultLen;
        else
            rto = (resultLen > rto) ? rto : resultLen;

        ostrstream stitle, squery;
        stitle << "Result: ";
        cexp.print(stitle);
        stitle << " (" << resultLen << " documents found. #" << param_from;
        stitle << "--#" << rto << ')' << '\0'; // terminate explicitly
        cexp.print(squery); squery << '\0';
        char *titlestr = stitle.str();
        char *querystr = squery.str();
        display_html(cout, "head.html", titlestr, querystr,
                     "(no error)", resultLen);
        print_results(descfile, result, uquery,
                      resultLen, param_n, param_from -1, rto - 1);
        display_html(cout, "foot.html", titlestr, querystr,
                     "(no error)", resultLen);
        writelog(uquery, resultLen);
        delete []titlestr;
        delete []querystr;
    }
    if(result) delete []result;
    
    return 0;
}
